<?php
/**
 * @file
 * 
 * Cache class for gathering heuristics.
 */

class HeuristicCache implements DrupalCacheInterface {
  static private $_prefetch = array();
  private $prefetch = array();
  private $backend = NULL;
  protected $bin;

  function clear($cid = NULL, $wildcard = FALSE) {
    return $this->backend->clear($cid, $wildcard);
  }

  function isEmpty() {
    return $this->backend->isEmpty();
  }

  function set($cid, $data, $expire = CACHE_PERMANENT) {
    return $this->backend->set($cid, $data, $expire);
  }

  function get($cid) {
    $cids = array($cid);
    $cache = $this->getMultiple($cids);
    return reset($cache);
  }

  /**
   * Constructor.
   * Initialize bin and prefetch data
   */
  function __construct($bin) {
    $this->bin = $bin;
    // Initialize prefetch for this bin
    if (!isset(self::$_prefetch[$bin])) {
      self::$_prefetch[$bin] = array();
    }
    $this->prefetch = &self::$_prefetch[$bin];

    // Let our parent do his work
    $class = variable_get('heuristic_cache_class_' . $bin, NULL);
    if (!isset($class)) {
      $class = variable_get('heuristic_cache_default_class', 'DrupalDatabaseCache');
    }
    $this->backend = new $class($bin);

    // Fetch all the cache now
    $this->prefetchCache();
  }

  /**
   * Aggregate cache data from prefetch and requested
   */
  function getMultiple(&$cids) {
    $fetch = array();
    $result = array();

    // Use prefetch data where available
    foreach ($cids as $cid) {
      if (isset($this->prefetch[$cid]) && is_object($this->prefetch[$cid])) {
        $result[$cid] = $this->prefetch[$cid];
        $this->prefetch[$cid] = FALSE;
      }
      else {
        $fetch[] = $cid;
      }
    }

    // Get all data that wasn't prefetched, from caching backend
    if ($fetch) {
      $fetched = $this->backend->getMultiple($fetch);
      $cids = $fetch;
      foreach (array_keys($fetched) as $cid) {
        $this->prefetch[$cid] = TRUE;
        $result[$cid] = &$fetched[$cid];
      }
    }
    else {
      $cids = array();
    }

    return $result;
  }

  /**
   * Prefetch cache
   */
  function prefetchCache() {
    $cid = 'hc:' . request_uri();
    $cids = array($cid);
    $cache = $this->backend->getMultiple($cids);
    if (isset($cache[$cid]) && is_array($cache[$cid]->data)) {
      $this->prefetch = $this->backend->getMultiple($cache[$cid]->data);
    }
    else {
      $this->prefetch = array();
    }

    // Save keys to prefetch next time.
    // Must use shutdown function. Destructor too late.
    register_shutdown_function(array($this, 'saveHeuristics'));
  }

  /**
   * Save keys to prefetch for this request
   */
  function saveHeuristics() {
    $update_necessary = FALSE;
    foreach ($this->prefetch as $cid => $cache) {
      if ($cache === FALSE) {
        $cids[] = $cid;
      }
      if ($cache === TRUE) {
        $cids[] = $cid;
        $update_necessary = TRUE;
      }
    }
    if ($update_necessary) {
      $cid = 'hc:' . request_uri();
      $this->backend->set($cid, $cids);
    }
  }
}

